#
# Copyright (C) 2013 Shang Yuanchun <idealities@gmail.com>
#
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# mirror is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mirror. If not, write to:
#   The Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor
#   Boston, MA  02110-1301, USA.
#
#

from mirror.sysinfo import enum

FOREGROUND_COLORS = enum(
                    BLACK  = '\033[30m',
                    RED    = '\033[31m',
                    GREEN  = '\033[32m',
                    YELLOW = '\033[33m',
                    BLUE   = '\033[34m',
                    MAGENTA= '\033[35m',
                    CYAN   = '\033[36m',
                    WHITE  = '\033[37m',
                    )

BACKGROUND_COLORS = enum(
                    BLACK  = '\033[40m',
                    RED    = '\033[41m',
                    GREEN  = '\033[42m',
                    YELLOW = '\033[43m',
                    BLUE   = '\033[44m',
                    MAGENTA= '\033[45m',
                    CYAN   = '\033[46m',
                    WHITE  = '\033[47m',
                    )

COLOR_END = '\033[0m'



#
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# mirror is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mirror. If not, write to:
# 	The Free Software Foundation, Inc.,
# 	51 Franklin Street, Fifth Floor
# 	Boston, MA  02110-1301, USA.
#
#


"""Main starting point for Mirror.  Contains the main() entry point."""

import os, sys
import signal
from   optparse import OptionParser

import mirror.log
import mirror.error
import mirror.console

def version_callback(option, opt_str, value, parser):
    print(os.path.basename(sys.argv[0]) + ": " + mirror.common.get_version())
    sys.exit(0)

def start_daemon():
    """Entry point for daemon script"""
    import mirror.common
    mirror.common.setup_translations()

    # Setup the argument parser
    parser = OptionParser(usage="%prog [options]")
    parser.add_option("-v", "--version", action="callback",
                      callback=version_callback,
                      help=_("Show program's version number and exit"))
    parser.add_option("-D", "--do-not-daemonize", dest="donot",
                      help=_("Do not daemonize (default is daemonize)"),
                      action="store_true",
                      default=False)
    parser.add_option("-c", "--config", dest="config",
                      help=_("Set the config location directory"),
                      action="store", type="str")
    parser.add_option("-P", "--pidfile", dest="pidfile",
                      help=_("Use pidfile to store process id"),
                      action="store", type="str")
    parser.add_option("-u", "--user", dest="user",
                      help=_("User to switch to. Need to start as root"),
                      action="store", type="str")
    parser.add_option("-g", "--group", dest="group",
                      help=_("Group to switch to. Need to start as root"),
                      action="store", type="str")
    parser.add_option("-l", "--logfile", dest="logfile",
                      help=_("Set the logfile location"),
                      action="store", type="str")
    parser.add_option("-L", "--loglevel", dest="loglevel",
                      help=_("Set the log level: none, info, warning, error, "
                             "critical, debug"),
                      action="store", type="str")
    parser.add_option("-q", "--quiet", dest="quiet",
                      help=_("Sets the log level to 'none', this is the same as `-L none`"),
                      action="store_true", default=False)
    parser.add_option("-r", "--rotate-logs",
                      help=_("Rotate logfiles."),
                      action="store_true", default=False)
    parser.add_option("--profile", dest="profile",
                      help=_("Profiles the daemon"),
                      action="store_true", default=False)
    parser.add_option("-t", "--tasks", dest="list_tasks",
                      help=_("List current tasks in scheduler's queue"),
                      action="store_true", default=False)
    parser.add_option("-s", "--signal", dest="signal",
                      help=_("Send signal to mirrord: stop, reload"),
                      action="store", type="str")

    # Get the options and args from the OptionParser
    (options, args) = parser.parse_args()

    if options.list_tasks:
        sys.exit(mirror.console.list_task_queue())

    if options.signal:
        sys.exit(mirror.console.signal_process(options.signal))

    if options.quiet:
        options.loglevel = "none"
    if not options.loglevel:
        options.loglevel = "info"

    logfile_mode = 'w'
    if options.rotate_logs:
        logfile_mode = 'a'

    import mirror.configmanager
    if options.config:
        if not mirror.configmanager.set_config_dir(options.config):
            print("There was an error setting the config dir! Exiting..")
            sys.exit(1)

    # Sets the options.logfile to point to the default location
    def set_logfile():
        if not options.logfile:
            options.logfile = os.path.join(mirror.common.DEFAULT_MIRRORD_LOG_DIR,
                                           "mirrord.log")

    set_logfile()

    # Setup the logger
    try:
        # Try to make the logfile's directory if it doesn't exist
        os.makedirs(os.path.abspath(os.path.dirname(options.logfile)))
    except:
        pass

    # Setup the logger
    if os.path.isfile(options.logfile):
        logfile_mode = 'a'
    mirror.log.setupLogger(level=options.loglevel,
                           filename=options.logfile,
                           filemode=logfile_mode)
    if options.donot:
        mirror.log.addStreamHandler(level=options.loglevel)

    # Writes out a pidfile if necessary
    def write_pidfile():
        if options.pidfile:
            open(options.pidfile, "wb").write("%s\n" % os.getpid())

    # If the do not daemonize is set, then we just skip the forking
    if not options.donot:
        if os.fork():
            # We've forked and this is now the parent process, so die!
            os._exit(0)
        os.setsid()
        # Do second fork
        if os.fork():
            os._exit(0)

    # Change to root directory
    os.chdir("/")
    # Write pid file before change gid and uid
    write_pidfile()

    if options.group:
        if not options.group.isdigit():
            import grp
            options.group = grp.getgrnam(options.group)[2]
        os.setgid(options.group)
    if options.user:
        if not options.user.isdigit():
            import pwd
            options.user = pwd.getpwnam(options.user)[2]
        os.setuid(options.user)

    # Close stdin, stdout, stderr ...
    if not options.donot:
        fp = open("/dev/null", 'r+')
        os.dup2(fp.fileno(), sys.stdin.fileno())
        os.dup2(fp.fileno(), sys.stdout.fileno())
        os.dup2(fp.fileno(), sys.stderr.fileno())
        fp.close()

    import logging
    log = logging.getLogger(__name__)

    try:
        mirror.common.check_mirrord_running(
            mirror.configmanager.get_config_dir("mirrord.pid"))
        # return fp to keep file not closed (by __exit__()), so the lock will not get released
        # we also write pid into it
        fp = mirror.common.lock_file(
            mirror.configmanager.get_config_dir("mirrord.pid"))
    except mirror.error.MirrordRunningError as e:
        log.error(e)
        log.error("You cannot run multiple daemons with the same config directory set.")
        sys.exit(1)
    except Exception as e:
        log.exception(e)
        sys.exit(1)

    import mirror.handler
    signal.signal(signal.SIGTERM, mirror.handler.shutdown_handler)
    signal.signal(signal.SIGQUIT, mirror.handler.shutdown_handler)
    signal.signal(signal.SIGINT,  mirror.handler.shutdown_handler)
    signal.signal(signal.SIGCHLD, mirror.handler.sigchld_handler)
    signal.signal(signal.SIGHUP,  mirror.handler.reload_handler)

    if options.profile:
        import hotshot
        hsp = hotshot.Profile(mirror.configmanager.get_config_dir("mirrord.profile"))
        hsp.start()
    try:
        log.info("Starting mirror daemon...")
        from mirror.daemon import MirrorDaemon
        daemon = MirrorDaemon(options, args)
    except Exception as e:
        log.exception(e)
        sys.exit(1)
    finally:
        if options.profile:
            hsp.stop()
            hsp.close()
            import hotshot.stats
            stats = hotshot.stats.load(mirror.configmanager.get_config_dir("mirrord.profile"))
            stats.strip_dirs()
            stats.sort_stats("time", "calls")
            stats.print_stats(400)



#
# Copyright (C) 2013 Shang Yuanchun <idealities@gmail.com>
#
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# mirror is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mirror. If not, write to:
#   The Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor
#   Boston, MA  02110-1301, USA.
#
#


import time
import logging
import threading
import mirror.component as component

log = logging.getLogger("pluginthread")

class PluginThread(threading.Thread):

    def __init__(self):
        threading.Thread.__init__(self, name="mirror.plugin")
        # python list is thread safe
        self.event_queue   = [ ]
        self.event_manager = component.get("EventManager")
        self.stop_event    = threading.Event()

    def add_event(self, event):
        self.event_queue.append(event)

    def run(self):
        log.debug("Plugin thread started")
        sleep_count = 0
        while (True):
            if self.stop_event.isSet() and len(self.event_queue) == 0:
                log.debug("PluginThread's stop_event is set, thread is exiting...")
                break
            try:
                event = self.event_queue.pop(0)
            except:
                time.sleep(0.1)
                sleep_count += 1
                if sleep_count > 600:
                    log.debug("Plugin thread finished")
                    break
                else:
                    continue
            # Call any handlers for the event
            for handler in self.event_manager.handlers[event.name]:
                log.debug("Running handler %s for event %s with args: %s",
                          event.name, handler, event.args)
                try:
                    handler(*event.args)
                except Exception as e:
                    log.error("Event handler %s failed in %s with exception: %s",
                              event.name, handler, e)




#
# Copyright (C) 2013 Shang Yuanchun <idealities@gmail.com>
#
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# mirror is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mirror. If not, write to:
#   The Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor
#   Boston, MA  02110-1301, USA.
#
#

import dbus
import dbus.service
import threading
from dbus.mainloop.glib import DBusGMainLoop

class MirrorBus(dbus.service.Object):
    def __init__(self, scheduler):
        self.scheduler = scheduler

        DBusGMainLoop(set_as_default=True)
        busname = dbus.service.BusName("cn.edu.bjtu.mirror", bus=dbus.SessionBus())
        dbus.service.Object.__init__(self, busname, "/cn/edu/bjtu/mirror")

    @dbus.service.method("cn.edu.bjtu.mirror")
    def list_queue(self):
        if not self.scheduler:
            return "No scheduler found, can not get task queue info."
        task_queue = self.scheduler.queue
        return str(task_queue)

    @dbus.service.method("cn.edu.bjtu.mirror")
    def active_task(self):
        if not self.scheduler:
            return "No scheduler found, can not get task queue info."
        return self.scheduler.active_tasks

    def start(self):
        import gobject
        loop = gobject.MainLoop()
        loop.run()

class MirrorBusThread(threading.Thread):
    def __init__(self, scheduler):
        threading.Thread.__init__(self, name="mirrorbus")
        self.mirrorbus = MirrorBus(scheduler)

    def run(self):
        self.mirrorbus.start()

if __name__ == "__main__":
    mirrorbus = MirrorBus(None)

    import gobject
    loop = gobject.MainLoop()
    loop.run()




#
# Copyright (C) 2013 Shang Yuanchun <idealities@gmail.com>
#
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# mirror is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mirror. If not, write to:
#   The Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor
#   Boston, MA  02110-1301, USA.
#
#

import os
import sys
import time
import shlex
import bisect
import signal
import logging
import mirror.common

from mirror.common import is_python3

log = logging.getLogger(__name__)

DEFAULT_ARGS = "--links --hard-links --times --verbose --delete --recursive"

PRIORITY_MIN = 1  # high priority
PRIORITY_MAX = 10 # low  priority

REGULAR_TASK = 1  # regular task
TIMEOUT_TASK = 2  # timeout checking task
SYSTEM_TASK  = 3  # system internal task, run in thread, so this is different
                  # with REGULAR_TASK

class AbstractTask(object):
    def __init__(self, name, scheduler_ref=None, **taskinfo):
        self.scheduler = (scheduler_ref() if scheduler_ref is not None else None)
        self.name      = name
        self.enabled   = True
        self.isinternal= False

        self.time      = taskinfo.get("time", None)
        if not self.time:
            log.error("Error in config for task: %s, time not set.", self.name)
            self.enabled = False
        crontime = mirror.common.parse_cron_time(self.time)
        if crontime:
            self.time_miniute = crontime[0]
            self.time_hour    = crontime[1]
            self.time_dom     = crontime[2]
            self.time_month   = crontime[3]
            self.time_dow     = crontime[4]
            for attr in ('time_miniute', 'time_hour', 'time_dom', 'time_month', 'time_dow'):
                if len(getattr(self, attr)) == 0:
                    log.error("Error in config for task: %s, time: %s not valid.",
                              self.name, attr)
                    self.enabled = False
        else:
            log.error("Error in config for task: %s, time not valid.", self.name)
            self.enabled = False

        try:
            self.priority = int(taskinfo['priority'])
            if self.priority < PRIORITY_MIN:
                self.priority = PRIORITY_MIN
            if self.priority > PRIORITY_MAX:
                self.priority = PRIORITY_MAX
        except:
            log.error("Error in config for task: %s, priority not valid.", self.name)
            self.priority = PRIORITY_MAX

        self.running = False
        if taskinfo.get("isinternal", False) != False:
            self.isinternal = True
            return

        if taskinfo.get("command", None) == None:
            log.warn("In config for task: %s, key: command not found, "
                     "using rsync as default",
                     name)
            command = "rsync"
        else:
            command = taskinfo.get("command")
        # NOTE: `command` is just a command name, but
        # self.command is the complete path...
        self.command   = mirror.common.find_command(command)
        self.cmdname   = command
        if not self.command:
            log.error("command `%s` not found in PATH, please install that first :)",
                      command)
            self.enabled = False
        self.pid       = 0
        self.stage     = 1
        self.code      = 0
        try:
            self.twostage = taskinfo['twostage'] != "0" 
            self.timeout  = mirror.common.parse_timeout(taskinfo['timeout'])
        except KeyError as e:
            log.error("Error in config for task: %s, key: %s not found.", self.name, e)
            self.enabled  = False

        if self.twostage and not taskinfo.has_key("firststage"):
            log.error("Error in config for task: %s, `twostage` is set but no `firststage`.", self.name)
            self.twostage = False
        if self.twostage:
            self.firststage = taskinfo['firststage']

    def run(self, stage = 1):
        try:
            self.stage = stage
            self.execute(stage)
        except Exception as e:
            log.error("Error occured when run `%s`: %s.", self.name, e)
            # If fork succeed but error occured before execv, we need to exit child process
            if os.getpid() == self.pid:
                sys.exit(1)
            # If we are in parent process, e.g. scheduler
            self.pid     = 0
            self.running = False

    def execute(self, stage):
        pid = os.fork()
        if pid > 0:
            self.pid        = pid
            self.running    = True
            self.start_time = int(time.time())
        elif pid == 0:
            self.pid   = os.getpid()
            if self.scheduler:
                logdir = self.scheduler.logdir
            else:
                logdir = mirror.common.DEFAULT_TASK_LOG_DIR
            if not os.path.exists(logdir):
                os.makedirs(logdir, 0o755)
            fp = open(logdir + self.name + '.log.' + time.strftime('%Y-%m-%d'), 'a')
            # Redirect child process's stdout and stderr
            os.dup2(fp.fileno(), sys.stdout.fileno())
            os.dup2(fp.fileno(), sys.stderr.fileno())
            fp.close()
            os.execv(self.command, self.get_args(stage))

    def stop(self, signo = signal.SIGTERM):
        if self.pid > 0:
            os.kill(self.pid, signo)
        if not self.twostage:
            self.set_stop_flag()

    def set_stop_flag(self):
        self.pid     = 0
        self.running = False

    TIME_STRUCT  = 1
    TIME_SECONDS = 2

    def get_schedule_time(self, since, style=TIME_SECONDS):
        """
        This method is too long.

        """

        if not hasattr(self, "time_miniute"):
            return None
        if not self.enabled:
            return None
        since_struct  = time.localtime(since)

        miniute = since_struct.tm_min
        hour    = since_struct.tm_hour
        day     = since_struct.tm_mday
        month   = since_struct.tm_mon
        year    = since_struct.tm_year

        day_increase   = False
        month_increase = False
        year_increase  = False
        if since_struct.tm_mon in self.time_month and since_struct.tm_mday in self.time_dom:
            miniute_idx = bisect.bisect(self.time_miniute, since_struct.tm_min)
            if since_struct.tm_hour in self.time_hour and miniute_idx < len(self.time_miniute):
                miniute = self.time_miniute[miniute_idx]
                hour    = since_struct.tm_hour
            else:
                miniute  = self.time_miniute[0]
                hour_idx = bisect.bisect(self.time_hour, since_struct.tm_hour)
                if hour_idx < len(self.time_hour):
                    hour = self.time_hour[hour_idx]
                else:
                    hour = self.time_hour[0]
                    day_increase = True
            if len(self.time_miniute) == 60 and len(self.time_hour) != 24:
                miniute  = self.time_miniute[0]
                hour_idx = bisect.bisect(self.time_hour, since_struct.tm_hour)
                if hour_idx < len(self.time_hour):
                    hour = self.time_hour[hour_idx]
                else:
                    hour = self.time_hour[0]
                    day_increase = True
            if not day_increase:
                day   = since_struct.tm_mday
                month = since_struct.tm_mon
        if since_struct.tm_mday not in self.time_dom or day_increase:
            miniute = self.time_miniute[0]
            hour    = self.time_hour[0]
            day_idx = bisect.bisect(self.time_dom, since_struct.tm_mday)
            if day_idx < len(self.time_dom):
                day = self.time_dom[day_idx]
            else:
                day = self.time_dom[0]
                month_increase = True
            if not month_increase:
                month = since_struct.tm_mon
        if since_struct.tm_mon not in self.time_month or month_increase:
            miniute = self.time_miniute[0]
            hour    = self.time_hour[0]
            day     = self.time_dom[0]
            month_idx = bisect.bisect(self.time_month, since_struct.tm_mon)
            if month_idx < len(self.time_month):
                month = self.time_month[month_idx]
            else:
                month = self.time_month[0]
                year_increase = True
        if year_increase:
            year += 1

        next_time   = time.mktime((year, month, day, hour, miniute, 0, 0, 0, 0))
        next_struct = time.localtime(next_time)
        if (next_struct.tm_wday + 1) not in self.time_dow:
            from datetime import datetime, timedelta
            wday_idx = bisect.bisect(self.time_dow, next_struct.tm_wday)
            if wday_idx < len(self.time_dow):
                wdays = self.time_dow[wday_idx] - next_struct.tm_wday
            else:
                wdays = (7 - self.time_dow[-1]) + self.time_dow[0]
            delta = timedelta(days = wdays)
            next_time = time.mktime((datetime.fromtimestamp(next_time) + delta).timetuple())

        if style == self.TIME_SECONDS:
            return next_time
        else:
            return time.localtime(next_time)

    def get_args(self, stage = 1):
        raise MirrorError("AbstractTask's get_args() is not implemented.")

class Task(AbstractTask):
    def __init__(self, name, scheduler_ref=None, **taskinfo):
        super(Task, self).__init__(name, scheduler_ref, **taskinfo)

        try:
            self.upstream = taskinfo['upstream[]']
            self.rsyncdir = taskinfo['rsyncdir']
            if self.rsyncdir[-1] != '/':
                self.rsyncdir += '/'
            self.localdir = taskinfo['localdir']
            if not os.path.exists(self.localdir):
                try:
                    os.makedirs(self.localdir, 0o755)
                except:
                    log.error("Error when create directory: %s", self.localdir)
        except KeyError as e:
            log.error("Error in config for mirror: %s, key: %s not found.", self.name, e)
            self.enabled  = False

        self.exclude  = taskinfo['exclude'] if taskinfo.has_key("exclude") else None
        self.args     = taskinfo['args']    if taskinfo.has_key("args")    else DEFAULT_ARGS

    def get_args(self, stage = 1):
        args  = [os.path.basename(self.command)]
        args += shlex.split(self.args)
        if self.twostage and stage == 2:
            args.remove("--delete")
            args.append("--delete-after")
        if self.exclude:
            args += shlex.split(self.exclude)
        if self.twostage and stage == 1:
            args += [self.upstream[0] + '::' + self.rsyncdir + self.firststage + '/',\
                     self.localdir + '/' + self.firststage]
            return args
        if self.twostage and stage == 2:
            args += ['--exclude', '/' + self.firststage + '/']
        args += [self.upstream[0] + '::' + self.rsyncdir,\
                 self.localdir]
        return args

class SimpleTask(AbstractTask):
    def __init__(self, name, scheduler_ref=None, **taskinfo):
        super(SimpleTask, self).__init__(name, scheduler_ref, **taskinfo)

        """
        If a SimpleTask is twostage, args is interpreted as args for second
        stage, firststage is interpreted as args for first stage.

        """
        self.args = taskinfo['args'] if taskinfo.has_key("args") else None

    def get_args(self, stage = 1):
        args  = [os.path.basename(self.command)]
        if self.twostage and stage == 1:
            args += self.firststage.split(" ")
            return args
        if self.args:
            args += shlex.split(self.args)
        return args

class SystemTask(AbstractTask):
    """
    This is internal task for mirror, so it will not fork.

    """
    def __init__(self, name, scheduler_ref=None, **taskinfo):
        if "isinternal" not in taskinfo:
            taskinfo["isinternal"] = True
        super(SystemTask, self).__init__(name, scheduler_ref, **taskinfo)

    def run(self):
        pass

    def stop(self):
        pass

# from name to Task
TASK_TYPES = { "simple" : SimpleTask }

if __name__ == "__main__":
    import mirror.log
    mirror.log.setupLogger(level="info",
                           filename="/home/ideal/.config/mirror/mirrord.log",
                           filemode="a")
    from mirror.configmanager import ConfigManager
    config = ConfigManager("mirror.ini")
    task   = Task('archlinux', None, **config['archlinux'])

    # The date to Shanghai
    since  = time.mktime((2013, 7, 20, 8, 0, 0, 0, 0, 0))
    print(time.ctime(task.get_schedule_time(since)))
    print(task.get_args())
    #task.run()
    #time.sleep(100)

    task   = Task('ubuntu', None, **config['ubuntu'])
    print(time.ctime(task.get_schedule_time(time.time())))
    print(task.get_args())
    print(task.get_args(stage=2))



#
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# mirror is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mirror. If not, write to:
# 	The Free Software Foundation, Inc.,
# 	51 Franklin Street, Fifth Floor
# 	Boston, MA  02110-1301, USA.
#
#

import os
import logging

import mirror.common
from   mirror.config import Config

log = logging.getLogger(__name__)

class _ConfigManager:
    def __init__(self):
        log.debug("ConfigManager started..")
        self.config_files = {}
        self.__config_directory = None

    @property
    def config_directory(self):
        if self.__config_directory is None:
            self.__config_directory = mirror.common.get_default_config_dir()
        return self.__config_directory

    def __del__(self):
        del self.config_files

    def set_config_dir(self, directory):
        """
        Sets the config directory.

        :param directory: str, the directory where the config info should be

        :returns bool: True if successfully changed directory, False if not
        """

        if not directory:
            return False

        log.info("Setting config directory to: %s", directory)
        if not os.path.exists(directory):
            # Try to create the config folder if it doesn't exist
            try:
                os.makedirs(directory)
            except Exception as e:
                log.error("Unable to make config directory: %s", e)
                return False
        elif not os.path.isdir(directory):
            log.error("Config directory needs to be a directory!")
            return False

        self.__config_directory = directory

        # Reset the config_files so we don't get config from old config folder
        # TODO: Probably should have it go through the config_files dict and try
        # to reload based on the new config directory
        self.save()
        self.config_files = {}

        return True

    def get_config_dir(self):
        return self.config_directory

    def close(self, config):
        """Close a config file."""
        try:
            del self.config_files[config]
        except KeyError:
            pass

    def save(self):
        """Save all the configs to disk."""
        for value in self.config_files.values():
            value.save()

        return True

    def get_config(self, config_file, need_reload=False, defaults=None):
        """Get a reference to the Config object for this filename"""
        log.debug("Getting config '%s'", config_file)
        # Create the config object if not already created
        if (need_reload or
            config_file not in self.config_files):
            self.config_files[config_file] = Config(config_file, defaults, self.config_directory)

        return self.config_files[config_file]

# Singleton functions
_configmanager = _ConfigManager()

# The parameter "config" is the name of config file
def ConfigManager(config, need_reload=False, defaults=None):
    return _configmanager.get_config(config, need_reload, defaults)

def set_config_dir(directory):
    """Sets the config directory, else just uses default"""
    return _configmanager.set_config_dir(directory)

def get_config_dir(filename=None):
    if filename != None:
        return os.path.join(_configmanager.get_config_dir(), filename)
    else:
        return _configmanager.get_config_dir()

def close(config):
    return _configmanager.close(config)



#
# Copyright (C) 2014 Shang Yuanchun <idealities@gmail.com>
#
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# mirror is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mirror. If not, write to:
#   The Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor
#   Boston, MA  02110-1301, USA.
#
#

import unittest
import time

from mirror.task import Task

class TaskTestCase(unittest.TestCase):

    def test_task(self):
        config = {
                 'upstream[]': 'mirror.bjtu.edu.cn',
                 'command': 'rsync',
                 'exclude': '',
                 'time':  '* */2 * * *',
                 'rsyncdir': 'archlinux/',
                 'localdir': '/home/mirror/archlinux',
                 'args': '--links --hard-links --times --verbose --delete --recursive',
                 'twostage': '0',
                 'timeout': '2h',
                 'propority': '2',
                 }
        task   = Task('archlinux', None, **config)
        since  = time.mktime((2013, 7, 20, 8, 0, 0, 0, 0, 0))
        self.assertEqual(time.ctime(task.get_schedule_time(since)),
                         'Sat Jul 20 10:00:00 2013')

if __name__ == '__main__':
    unittest.main()



#
# Copyright (C) 2013 Shang Yuanchun <idealities@gmail.com>
#
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# mirror is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mirror. If not, write to:
#   The Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor
#   Boston, MA  02110-1301, USA.
#
#


import logging
from mirror.component import Component

log = logging.getLogger(__name__)

class PluginBase(Component):

    def __init__(self, name):
        super(PluginBase, self).__init__(name)
        self.name = name
        log.debug("Plugin initialized...")

    def enable(self):
        raise NotImplementedError("Need to define an enable method!")

    def disable(self):
        raise NotImplementedError("Need to define a disable method!")




#
# Copyright (C) 2013 Shang Yuanchun <idealities@gmail.com>
#
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# mirror is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mirror. If not, write to:
#   The Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor
#   Boston, MA  02110-1301, USA.
#
#


import os
import logging
import pkg_resources
import mirror.common
import mirror.configmanager
import mirror.component as component
from   mirror.component import Component

log = logging.getLogger("pluginmanager")

METADATA_KEYS = (
    "Name",
    "License",
    "Author",
    "Home-page",
    "Summary",
    "Platform",
    "Version",
    "Author-email",
    "Description",
)

class PluginManager(Component):

    def __init__(self, config_file, entry_name):
        super(PluginManager, self).__init__(PluginManager.__name__)
        self.config     = mirror.configmanager.ConfigManager(config_file)
        self.entry_name = entry_name

        # available plugins
        self.available_plugins = []

        # enabled plugins
        self.plugins    = {}
        self.hooks      = {}

        self.scan_plugins()

    def enable_plugins(self):
        for name in self.available_plugins:
            self.enable_plugin(name)

    def enable_plugin(self, plugin_name):
        if plugin_name in self.plugins:
            log.warn("Cannot enable an already enabled plugin %s", plugin_name)
            return

        plugin_name = plugin_name.replace(' ', '-')
        egg         = self.pkg_env[plugin_name][0]
        egg.activate()
        for name in egg.get_entry_map(self.entry_name):
            entry_point = egg.get_entry_info(self.entry_name, name)
            try:
                cls      = entry_point.load()
                if not cls.enabled:
                    continue
                instance = cls(plugin_name.replace('-', '_'))
            except Exception as e:
                log.error("Unable to instantiate plugin %r from %r!",
                          name, egg.location)
                log.exception(e)
                continue
            instance.enable()
            if not instance.__module__.startswith("mirror.plugins."):
                log.warn("Wrong module for plugin: %s", name)

            component.start([instance.plugin.name])
            plugin_name = plugin_name.replace('-', ' ')
            self.plugins[plugin_name] = instance
            log.info("Plugin %s enabled...", plugin_name)

    def disable_plugins(self):
        plugin_names = self.plugins.keys()
        for name in plugin_names:
            self.disable_plugin(name)

    def disable_plugin(self, plugin_name):
        if plugin_name not in self.plugins:
            log.info("Plugin %s is not enabled, no need to disable", plugin_name)
            return

        instance = self.plugins[plugin_name]
        instance.disable()
        component.deregister(instance)
        self.plugins.pop(plugin_name)

    def scan_plugins(self):
        """
        Scans for available plugins

        """
        base_plugin_dir = mirror.common.resource_filename("mirror", "plugins")
        pkg_resources.working_set.add_entry(base_plugin_dir)
        user_plugin_dir = os.path.join(mirror.configmanager.get_config_dir(), "plugins")

        plugins_dirs    = [ base_plugin_dir ]
        for dirname in os.listdir(base_plugin_dir):
            plugin_dir  = os.path.join(base_plugin_dir, dirname)
            pkg_resources.working_set.add_entry(plugin_dir)
            plugins_dirs.append(plugin_dir)

        pkg_resources.working_set.add_entry(user_plugin_dir)
        plugins_dirs.append(user_plugin_dir)

        self.pkg_env    = pkg_resources.Environment(plugins_dirs)

        for name in self.pkg_env:
            log.debug("Found plugin: %s %s at %s",
                       self.pkg_env[name][0].project_name,
                       self.pkg_env[name][0].version,
                       self.pkg_env[name][0].location)
            self.available_plugins.append(self.pkg_env[name][0].project_name)

    def __getitem__(self, key):
        return self.plugins[key]

    def get_available_plugins(self):
        return self.available_plugins

    def get_enabled_plugins(self):
        return self.plugins.keys()

    def get_plugin_info(self, name):
        info = dict.fromkeys(METADATA_KEYS)
        last_header = ""
        cont_lines  = []
        for line in self.pkg_env[name][0].get_metadata("PKG-INFO").splitlines():
            if not line:
                continue
            if line[0] in ' \t' and (
                len(line.split(":", 1)) == 1 or line.split(":", 1)[0] not in info):
                # This is a continuation
                cont_lines.append(line.strip())
            else:
                if cont_lines:
                    info[last_header] = "\n".join(cont_lines).strip()
                    cont_lines = []
                if line.split(":", 1)[0] in info:
                    last_header = line.split(":", 1)[0]
                    info[last_header] = line.split(":", 1)[1].strip()

        return info

    def start(self):
        self.enable_plugins()

    def stop(self):
        self.disable_plugins()




#
# Copyright (C) 2013 Shang Yuanchun <idealities@gmail.com>
#
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# mirror is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mirror. If not, write to:
#   The Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor
#   Boston, MA  02110-1301, USA.
#
#

import os
import time
import signal
import logging
import mirror.component as component

log = logging.getLogger(__name__)

signals = { getattr(signal, sigtxt) : sigtxt
            for sigtxt in dir(signal) if sigtxt.startswith("SIG") and
            not sigtxt.startswith("SIG_") }

def shutdown_handler(signo, frame):
    import mirror.configmanager
    pidfile = mirror.configmanager.get_config_dir("mirrord.pid")
    if os.path.isfile(pidfile):
        os.remove(pidfile)
    log.info("Got signal %s, exiting...", signals[signo])

    import sys
    scheduler = component.get("Scheduler")
    if scheduler is None:
        sys.exit(0)

    # We will waitpid() in stop_all_tasks(),
    # so unregister sigchld_handler here.
    signal.signal(signal.SIGCHLD, signal.SIG_DFL)
    scheduler.stop_all_tasks()

    if hasattr(scheduler, "buspid"):
        os.kill(scheduler.buspid, signal.SIGTERM)
        pid, status = os.waitpid(scheduler.buspid, 0)
        log.info("Killed mirror dbus with pid: %d", pid)

    event_manager = component.get("EventManager")
    while (event_manager.plugin_thread and
           len(event_manager.plugin_thread.event_queue) > 0):
        time.sleep(0.1)
    component.deregister(event_manager)
    component.deregister(component.get("PluginManager"))

    # Deregister the scheduler,
    # this will call scheduler's stop().
    # But scheduler's start() is called
    # directly, not by component's start().
    component.deregister(scheduler)

    log.info("Bye bye... :)")
    sys.exit(0)

def sigchld_handler(signo, frame):
    try:
        pid, status = os.waitpid(-1, os.WNOHANG)
    except OSError as e:
        log.error("Error occured when waitpid(), %s.", e)
        return
    scheduler = component.get("Scheduler")
    if scheduler is None:
        return
    scheduler.stop_task_with_pid(pid, status)

def reload_handler(signo, frame):
    log.info("Got signal %s, start reloading...", signals[signo])
    scheduler = component.get("Scheduler")
    scheduler.reload_config()



#
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# mirror is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mirror. If not, write to:
# 	The Free Software Foundation, Inc.,
# 	51 Franklin Street, Fifth Floor
# 	Boston, MA  02110-1301, USA.
#
#

"""Logging functions"""

import logging

from mirror import common

__all__ = ["setupLogger", "setLoggerLevel", "addStreamHandler"]

if 'dev' in common.get_version():
    DEFAULT_LOGGING_FORMAT = "%%(asctime)s.%%(msecs)03.0f [%%(levelname)-8s]"\
                             "[%%(name)-%ds:%%(lineno)-4d] %%(message)s"
else:
    DEFAULT_LOGGING_FORMAT = "%%(asctime)s [%%(levelname)-8s][%%(name)-%ds] %%(message)s"

MAX_LOGGER_NAME_LENGTH = 16

levels = {
    "none":     logging.NOTSET,
    "info":     logging.INFO,
    "warn":     logging.WARNING,
    "warning":  logging.WARNING,
    "error":    logging.ERROR,
    "critical": logging.CRITICAL,
    "debug":    logging.DEBUG,
    "trace":    5,
    "garbage":  1
}

def setupLogger(level="error", filename=None, filemode="w"):
    """
    Sets up the basic logger and if `:param:filename` is set, then it will log
    to that file instead of stdout.

    :param level: str, the level to log
    :param filename: str, the file to log to
    """
    import logging
    logging.setLoggerClass(logging.Logger)
    logging.addLevelName(5, 'TRACE')
    logging.addLevelName(1, 'GARBAGE')

    level = levels.get(level, logging.ERROR)

    rootLogger = logging.getLogger()

    if filename and filemode=='a':
        import logging.handlers
        handler = logging.handlers.RotatingFileHandler(
            filename, filemode,
            maxBytes=50*1024*1024,   # 50 Mb
            backupCount=8,
            encoding='utf-8',
            delay=0
        )
    elif filename and filemode=='w':
        import logging.handlers
        handler = getattr(
            logging.handlers, 'WatchedFileHandler', logging.FileHandler)(
                filename, filemode, 'utf-8', delay=0
        )
    else:
        handler = logging.StreamHandler()

    handler.setLevel(level)

    formatter = logging.Formatter(
        DEFAULT_LOGGING_FORMAT % MAX_LOGGER_NAME_LENGTH,
        datefmt="%m-%d %H:%M:%S"
    )

    handler.setFormatter(formatter)

    rootLogger.addHandler(handler)
    rootLogger.setLevel(level)

def setLoggerLevel(level, logger_name=None):
    """
    Sets the logger level.

    :param level: str, a string representing the desired level
    :param logger_name: str, a string representing desired logger name for which
                        the level should change. The default is "None" will will
                        tweak the root logger level.

    """
    logging.getLogger(logger_name).setLevel(levels.get(level, logging.ERROR))

def addStreamHandler(level="error", force=False):
    """
    Add a stream handler to logger. If there is already one and force is not set, the addition is canceled.

    :param level: str, a string representing the desired level
    :param force: bool, whether to add when there is already one

    """
    if not force:
        for handler in logging.getLogger().handlers:
            if handler.__class__ == logging.StreamHandler:
                return

    level = levels.get(level, logging.ERROR)

    handler = logging.StreamHandler()
    handler.setLevel(level)

    formatter = logging.Formatter(
        DEFAULT_LOGGING_FORMAT % MAX_LOGGER_NAME_LENGTH,
        datefmt="%m-%d %H:%M:%S"
    )

    handler.setFormatter(formatter)

    logging.getLogger().addHandler(handler)



#
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# mirror is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mirror. If not, write to:
#   The Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor
#   Boston, MA  02110-1301, USA.
#
#

import logging
from   mirror.component import Component
from   mirror.pluginthread import PluginThread

log = logging.getLogger(__name__)

class EventManager(Component):

    def __init__(self):
        # The name is "EventManager"
        super(EventManager, self).__init__(self.__class__.__name__)

        self.handlers = {}
        self.plugin_thread = None

    def emit(self, event):
        """
        Emits the event to interested clients.

        :param event: MirrorEvent
        """
        if event.name not in self.handlers:
            return

        if (not self.plugin_thread) or (not self.plugin_thread.isAlive()):
            if self.plugin_thread:
                del self.plugin_thread
            self.plugin_thread = PluginThread()
            self.plugin_thread.start()
        self.plugin_thread.add_event(event)

    def register_event_handler(self, event, handler):
        """
        Register a function to be called when a `:param:event` is emitted.

        :param event: string, the event name
        :param handler: function, to be called when `:param:event` is emitted

        """
        if event not in self.handlers:
            self.handlers[event] = []

        if handler not in self.handlers[event]:
            self.handlers[event].append(handler)

    def deregister_event_handler(self, event, handler):
        """
        Deregisters an event handler function.

        :param event: string, the event name
        :param handler: function, currently registered to handle `:param:event`

        """
        if event in self.handlers and handler in self.handlers[event]:
            self.handlers[event].remove(handler)

    def stop(self):
        """
        When EventManager is stopped, if plugin_thread is still alive,
        set the stop_event to make it exit.

        """
        if self.plugin_thread and self.plugin_thread.isAlive():
            self.plugin_thread.stop_event.set()



#
# Copyright (C) 2013 Shang Yuanchun <idealities@gmail.com>
#
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# mirror is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mirror. If not, write to:
#   The Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor
#   Boston, MA  02110-1301, USA.
#
#

import bisect

class TaskInfo():
    def __init__(self, name, tasktype, time, priority):
        self.name     = name
        self.tasktype = tasktype
        self.time     = time
        self.priority = priority

    def __eq__(self, other):
        return self.name == other.name

    def __cmp__(self, other):
        return (self.time - other.time) if self.time != other.time else (
                                           self.priority - other.priority)

    def __str__(self):
        return "(name: %s, type: %d, time: %d, priority: %d)" % (self.name,
                                    self.tasktype, self.time, self.priority)

class Queue(object):
    def __init__(self, *items):
        self.queue = []
        for item in items:
            if not isinstance(item, TaskInfo):
                continue
            self.put(item)

    def put(self, item):
        bisect.insort(self.queue, item)

    def get(self):
        try:
            return self.queue.pop(0)
        except:
            return None

    def remove(self, item):
        self.queue.remove(item)

    def empty(self):
        return ( len(self.queue) == 0 )

    def size(self, key = None, value = None):
        if key is None or value is None:
            return len(self.queue)
        else:
            if self.empty():
                return 0
            if not hasattr(self.queue[0], key):
                return 0
            count = 0
            for item in self.queue:
                count += ( getattr(item, key) == value )
            return count

    def find(self, name):
        for taskinfo in self.queue:
            if taskinfo.name == name:
                return taskinfo
        return None

    def __getitem__(self, key):
        try:
            return self.queue[key]
        except:
            return None

    def __str__(self):
        return "[" + ",".join([str(item) for item in self.queue]) + "]"

    def __iter__(self):
        for item in self.queue:
            yield item

    def __contains__(self, item):
        return item in self.queue

if __name__ == "__main__":
    task1 = TaskInfo("Buy clock",    0, 1376712000, 2)
    task2 = TaskInfo("Basketball",   0, 1376701200, 1)
    task3 = TaskInfo("Wash clothes", 0, 1376701200, 2)
    task4 = TaskInfo("Send letter",  0, 1376704800, 4)

    queue = Queue(task1, task2, task3, task4, "Qiandao Lake")
    assert(queue.size() == 4)
    assert(not queue.empty())
    print(queue)

    task  = queue[0]
    print(task)

    print(task in queue)
    task4 = TaskInfo("Send letter",  0, 1376704800, 4)
    print(task4 in queue)

    print(queue.size("tasktype", 0))




#
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# mirror is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mirror. If not, write to:
#   The Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor
#   Boston, MA  02110-1301, USA.
#
#


import os
import re
import socket

d = { 1: 0, 5: 1, 15: 2}

def loadavg(duration = 5):
    if duration not in d:
        duration = 5
    if hasattr(os, "getloadavg"):
        return os.getloadavg()[d[duration]]
    try:
        loadavgs = open("/proc/loadavg").read().split(" ")
        return float(loadavgs[d[duration]])
    except:
        return 0.0

def enum(**enums):
    return type('Enum', (), enums)

# st in enum in /usr/include/netinet/tcp.h
TCP_STATUS = enum(TCP_ESTABLISHED =1,\
                  TCP_SYN_SENT    =2,\
                  TCP_SYN_RECV    =3,\
                  TCP_FIN_WAIT1   =4,\
                  TCP_FIN_WAIT2   =5,\
                  TCP_TIME_WAIT   =6,\
                  TCP_CLOSE       =7,\
                  TCP_CLOSE_WAIT  =8,\
                  TCP_LAST_ACK    =9,\
                  TCP_LISTEN      =10,\
                  TCP_CLOSING     =11)

pattern = (r"\d+:\s+" + 
           r"(?P<local_addr>[\da-fA-F]+):(?P<local_port>[\da-fA-F]+)\s+" +
           r"(?P<remote_addr>[\da-fA-F]+):(?P<remote_port>[\da-fA-F]+)\s+" +
           r"(?P<status>[\da-fA-F]+)\s+[\da-fA-F]+:[\da-fA-F]+\s+[\da-fA-F]+:[\da-fA-F]+\s+[\da-fA-F]+\s+\d+\s+\d+\s+" +
           r"(?P<inode>\d+)")

def tcpconn(port = 80):
    try:
        # TODO: using AF_NETLINK to fetch tcp infomation instead from /proc
        sock = socket.socket(socket.AF_NETLINK, socket.SOCK_RAW)
        sock.close()
    except Exception as e:
        pass
    connections = 0
    tcp   = re.compile(pattern)
    files = ("/proc/net/tcp", "/proc/net/tcp6")
    for path in files:
        if os.access(path, os.R_OK):
            fp = open(path)
            fp.readline() # skip title
            for line in fp:
                conn = tcp.search(line).groupdict()
                local_port = int(conn['local_port'], 16)
                if local_port != port:
                    continue
                status = int(conn['status'], 16)
                if status == TCP_STATUS.TCP_ESTABLISHED:
                    connections += 1
            fp.close()
    return connections

if __name__ == "__main__":
    print("Current connections: %d" % tcpconn(port = 44971))



#
# Copyright (C) 2013 Shang Yuanchun <idealities@gmail.com>
#
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# mirror is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mirror. If not, write to:
#   The Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor
#   Boston, MA  02110-1301, USA.
#
#


"""Scheduler for Mirror :("""

import os, sys
import time
import fcntl
import signal
import logging
import weakref
import mmap
import struct
import cPickle as pickle
import mirror.common
import mirror.error
import mirror.handler
import mirror.event
import mirror.component   as     component
from mirror.configmanager import ConfigManager
from mirror.task          import Task, SimpleTask, TASK_TYPES
from mirror.task          import PRIORITY_MIN, PRIORITY_MAX
from mirror.task          import REGULAR_TASK, TIMEOUT_TASK, SYSTEM_TASK
from mirror.sysinfo       import loadavg, tcpconn
from mirror.queue         import TaskInfo, Queue
from mirror.component     import Component

from collections import OrderedDict as odict

log = logging.getLogger(__name__)

class Scheduler(Component):
    CHECK_TIMEOUT = 0x01
    SCHEDULE_TASK = 0x02

    def __init__(self, options = None, args = None):
        # The name is "Scheduler"
        super(Scheduler, self).__init__(self.__class__.__name__)
        # self.tasks contains all tasks needed to run in theory,
        # including the tasks that are not enabled
        self.config  = ConfigManager("mirror.ini")
        self.tasks   = odict()
        self.queue   = Queue()
        self.todo    = self.SCHEDULE_TASK
        # the number of tasks that enabled
        self.active_tasks    = -1
        self.roused_by_child = False

        self.init_general(self.config)
        self.init_tasks  (self.config)

    def start(self):
        event_manager = component.get("EventManager")
        event_manager.emit(mirror.event.MirrorStartEvent())
        while (True):
            self.sleep()
            if not self.roused_by_child:
                log.info("I am waking up...")
            self.schedule()

    TODO = { REGULAR_TASK : SCHEDULE_TASK,
             SYSTEM_TASK  : SCHEDULE_TASK,
             TIMEOUT_TASK : CHECK_TIMEOUT,
            }

    def sleep(self):
        self.append_tasks()
        self.write_mmap()

        nexttask  = self.queue[0]
        self.todo = 0
        if nexttask:
            for taskinfo in self.queue:
                if taskinfo.time > nexttask.time:
                    break
                self.todo |= self.TODO.get(taskinfo.tasktype, 0)
        # nexttask.time - time.time() is
        # the duration we can sleep...
        if nexttask:
            sleeptime = nexttask.time - time.time()
        else:
            sleeptime = 1800 # half an hour
        log.info("I am going to sleep, next waking up: %s",
                 time.ctime(time.time() + sleeptime))
        self.roused_by_child = False
        time.sleep(sleeptime)

    def schedule(self):
        if self.queue.empty():
            log.info("But no task needed to start...")
            return

        self.init_sysinfo()

        # we do not need microseconds
        curtime    = int(time.time())
        taskqueue  = [ taskinfo for taskinfo in self.queue ]
        if ( self.todo & self.SCHEDULE_TASK ):
            # to move to zero second
            timestamp  = curtime
            timestamp -= curtime   % 60
            # next miniute
            end        = timestamp + 60
            for taskinfo in taskqueue:
                if self.TODO.get(taskinfo.tasktype, 0) != self.SCHEDULE_TASK:
                    continue
                if taskinfo.time < timestamp:
                    log.info("Strange problem happened,"
                             "task: %s schedule time is in past,"
                             "maybe I sleeped too long...", taskinfo.name)
                    self.queue.remove(taskinfo)
                    self.append_task(taskinfo.name, self.tasks[taskinfo.name], since = end)
                if taskinfo.time >= end:
                    break
                if taskinfo.time >= timestamp and taskinfo.time < end:
                    self.schedule_task(taskinfo)

        if ( self.todo & self.CHECK_TIMEOUT ):
            for taskinfo in taskqueue:
                if self.TODO.get(taskinfo.tasktype, 0) != self.CHECK_TIMEOUT:
                    continue
                if taskinfo.time <= curtime:
                    log.info("Task: %s timeouts",
                             taskinfo.name)
                    self.stop_task(taskinfo)
                if taskinfo.time > curtime:
                    break

    def schedule_task(self, taskinfo):
        """
        Schedule a task, but it is not guaranteed that it will really be run, it is 
        decided by some conditions, e.g. system load, current http connections.

        NOTE:
        The priority that a task can run is a function of ( current value / limit ).
        See get_runnable_priority(). However an exception is `maxtasks`, if current
        running tasks is reaching `maxtasks`, only specific priority (lower than or
        equal to 2) tasks can still be running.

        """
        task = self.tasks[taskinfo.name]
        if task.isinternal:
            self.run_system_task(taskinfo)
            return

        if task.priority > self.get_runnable_priority(self.current_load, self.loadlimit):
            log.info("Task: %s not scheduled because system load %.2f is too high",
                     taskinfo.name, self.current_load)
            self.delay_task(taskinfo)
            return
        if task.priority > self.get_runnable_priority(self.current_conn,  self.httpconn):
            log.info("Task: %s not scheduled because http connections is too many",
                     taskinfo.name)
            self.delay_task(taskinfo)
            return
        if self.maxtasks > 0 and self.count_running_tasks() >= self.maxtasks and task.priority > 2:
            log.info("Task: %s not scheduled because running tasks is larger than %d",
                     taskinfo, self.maxtasks)
            self.delay_task(taskinfo)
            return
        log.info("Starting task: %s ...", taskinfo.name)
        self.run_task(taskinfo)

    def init_sysinfo(self):
        """
        Get system info for this turn of schedule().

        """
        self.current_load = loadavg()
        self.current_conn = tcpconn()

    def delay_task(self, taskinfo, delay_seconds = 1800):
        """
        If a task is not scheduled due to some reason, it will be
        delayed for `delay_seconds` seconds (wich is default half
        an hour), when task's next schedule time is later than that,
        else it's set to task's next schedule time.

        """
        if taskinfo not in self.queue:
            return
        task      = self.tasks[taskinfo.name]
        next_time = task.get_schedule_time(since = time.time())
        if taskinfo.time + delay_seconds > next_time:
            taskinfo.time  = next_time
        else:
            # In python objects are passed by reference
            taskinfo.time += delay_seconds
        self.reappend_task(task, taskinfo)

    def count_running_tasks(self):
        """
        Calculate the number of current running tasks.

        """
        running = 0
        for taskname, task in self.tasks.iteritems():
            if task.isinternal:
                continue
            running += task.running
        return running

    def append_tasks(self):
        """
        Append the tasks that are needed to run into self.queue.

        NOTE:
        If a task is currently running or it is not enabled, it will
        not be added to the queue.

        """
        now = time.time()
        for taskname in self.tasks:
            task = self.tasks[taskname]
            self.append_task(taskname, task, since = now)

    def append_task(self, taskname, task, since):
        """
        In some cases a task with same name may be ignored if there
        is a running one, but this is a feature, not a bug...

        """
        if task.running and task.timeout > 0:
            return
        if not task.enabled:
            return
        taskinfo = TaskInfo(taskname, (SYSTEM_TASK if task.isinternal else REGULAR_TASK),
                            task.get_schedule_time(since), task.priority)

        # however this is hard to understand,
        # for system tasks and timeout tasks
        if taskinfo in self.queue:
            return
        self.queue.put(taskinfo)
        event_manager = component.get("EventManager")
        event_manager.emit(mirror.event.TaskEnqueueEvent(taskname))

    def reappend_task(self, task, taskinfo):
        """
        Remove a taskinfo from queue and put it in again,
        to keep the queue in order.

        """
        if taskinfo not in self.queue:
            return
        self.queue.remove(taskinfo)
        self.queue.put(taskinfo)
        event_manager = component.get("EventManager")
        event_manager.emit(mirror.event.TaskEnqueueEvent(taskinfo.name))

    DEFAULT_BUFFER_SIZE  = 10240

    def write_mmap(self):
        data = pickle.dumps(self.queue)
        size = len(data) + 2 + 4
        if not hasattr(self, "buffer") or self.buffersz < size:
            self.buffersz = max(self.DEFAULT_BUFFER_SIZE, size)
            if hasattr(self, "buffer"):
                self.buffer.close()
            self.bufferfd = os.open("/tmp/mirrord",
                                    os.O_CREAT | os.O_TRUNC | os.O_RDWR,
                                    0o644)
            flag = fcntl.fcntl(self.bufferfd, fcntl.F_GETFD)
            fcntl.fcntl(self.bufferfd, fcntl.F_SETFD, flag | fcntl.FD_CLOEXEC)
            os.write(self.bufferfd, '\x00' * self.buffersz)
            self.buffer   = mmap.mmap(self.bufferfd, self.buffersz, mmap.MAP_SHARED, mmap.PROT_WRITE)
            # close bufferfd
            os.close(self.bufferfd)
            self.buffer.write("\x79\x71")
        self.buffer.seek(2)
        self.buffer.write(struct.pack("I", size))
        self.buffer.write(data)

    def stop(self):
        log.info("Stopping mirror scheduler")
        if not hasattr(self, "buffer"):
            return
        self.buffer.close()
        os.unlink("/tmp/mirrord")

    def append_timeout_task(self, taskname, task, time):
        """
        A timeout checking task is added after a task begins to run.

        """
        if not task.running:
            return
        if not task.enabled:
            return
        taskinfo = TaskInfo(taskname, TIMEOUT_TASK,
                            time, task.priority)
        if taskinfo in self.queue:
            return
        self.queue.put(taskinfo)
        event_manager = component.get("EventManager")
        event_manager.emit(mirror.event.TaskEnqueueEvent(taskname))

    def remove_timeout_task(self, taskname):
        """
        It's slow...

        """
        taskqueue  = [ taskinfo for taskinfo in self.queue ]
        for taskinfo in taskqueue:
            if taskinfo.name == taskname and taskinfo.tasktype == TIMEOUT_TASK:
                self.queue.remove(taskinfo)
                return

    def init_general(self, config):
        self.emails    = []
        self.loadlimit = 4.0
        self.httpconn  = 1200
        self.logdir    = mirror.common.DEFAULT_TASK_LOG_DIR
        self.maxtasks  = 10

        if "general" not in config:
            log.error("Error in config file, no `general` section, will use default setting.")
            return
        import re
        emails = re.compile(r"([^@\s]+@[^@\s,]+)")
        emails = emails.findall(config['general']['emails'])
        for email in emails:
            self.emails.append(email)

        self.loadlimit = float(config['general']['loadlimit'])
        self.httpconn  = int  (config['general']['httpconn'] )
        self.maxtasks  = int  (config['general']['maxtasks'] )
        self.logdir    = config['general']['logdir']
        if self.logdir[-1] != os.path.sep:
            self.logdir += os.path.sep

    def reload_config(self):
        log.info("Stopping running tasks...")
        signal.signal(signal.SIGCHLD, signal.SIG_DFL)
        self.stop_all_tasks()
        signal.signal(signal.SIGCHLD, mirror.handler.sigchld_handler)

        log.info("Clearing old data...")
        for taskname, task in self.tasks.iteritems():
            if not task.isinternal:
                del self.tasks[taskname]
        self.queue = Queue()

        log.info("Reloading new configs...")
        self.config = ConfigManager("mirror.ini", need_reload = True)
        self.init_general(self.config)
        self.init_tasks  (self.config)

    def init_tasks(self, config):
        for mirror in config:
            if mirror == 'general':
                continue
            # We think it's default mirror.task.Task
            task_class = TASK_TYPES.get(config[mirror].get("type", None), Task)
            self.tasks[mirror] = task_class(mirror, weakref.ref(self), **config[mirror])
        self.active_tasks = len(
                            [mirror for mirror, task in self.tasks.iteritems() if task.enabled])

    def run_system_task(self, taskinfo):
        event_manager = component.get("EventManager")
        event_manager.emit(mirror.event.RunSystemTaskEvent(taskinfo))
        # after we run the system task, we need to update the queue
        # or else the sleeptime will be invalid
        if taskinfo in self.queue:
            self.queue.remove(taskinfo)
        task = self.tasks[taskinfo.name]
        self.append_task(taskinfo.name, task, time.time())

    def run_task(self, taskinfo, stage = 1):
        if taskinfo.name not in self.tasks:
            return
        task = self.tasks[taskinfo.name]
        # for tasks that is still running when next schedule time
        # is reached (but has no timeout set), we just need to
        # reappend it.
        if task.running and task.timeout <= 0:
            taskinfo.time  = task.get_schedule_time(since = time.time())
            self.reappend_task(task, taskinfo)
        if task.running and ( not task.twostage ):
            log.info("Task: %s is still running and no timeout set, skipped", taskinfo.name)
            return

        event_manager = component.get("EventManager")
        event_manager.emit(mirror.event.PreTaskStartEvent(taskinfo.name))
        task.run(stage)
        if taskinfo in self.queue:
            self.queue.remove(taskinfo)
        log.info("Task: %s begin to run with pid %d", taskinfo.name, task.pid)
        event_manager.emit(mirror.event.TaskStartEvent(taskinfo.name, task.pid))

        if task.timeout <= 0:
            self.append_task(taskinfo.name, task, time.time())
        else:
            self.append_timeout_task(taskinfo.name, task,
                                     task.start_time + task.timeout)

    def stop_task(self, taskinfo):
        """
        Stop a task, it should only be called when that task timeouts.

        """
        if taskinfo.name not in self.tasks:
            return
        task = self.tasks[taskinfo.name]
        if not task.running:
            return
        pid  = task.pid
        # Python's SIGCHLD sometimes has delay in calling its handler,
        # we have to disable sigchld_handler() here.
        # More: http://utcc.utoronto.ca/~cks/space/blog/python/CPythonSignals
        signal.signal(signal.SIGCHLD, signal.SIG_DFL)
        task.stop()
        self.stop_task_manually(task, pid)
        signal.signal(signal.SIGCHLD, mirror.handler.sigchld_handler)

    def stop_task_manually(self, task, pid):
        """
        Without SIGCHLD handler, we have to waitpid() here.

        """
        pid, status  = os.waitpid(pid, 0)
        endstr, code = self.parse_return_status(status)
        task.code    = code
        log.info("Killed task: %s %s %d, pid %d", task.name, endstr, code, pid)
        self.remove_timeout_task(task.name)
        self.task_post_process(task)

    def stop_task_with_pid(self, pid, status):
        """
        Change task's running and pid attr as it's stopped.

        """
        self.roused_by_child = True
        for taskname, task in self.tasks.iteritems():
            if task.isinternal:
                continue
            if task.pid == pid:
                if not task.running:
                    return
                endstr, code = self.parse_return_status(status)
                task.code    = code
                log.info("Task: %s %s %d, pid %d", taskname, endstr, code, pid)
                self.remove_timeout_task(taskname)
                self.task_post_process(task)
                return

    def task_post_process(self, task):
        """
        Check whether a task needs post process, e.g. two stage tasks.

        """
        event_manager = component.get("EventManager")
        if not task.twostage:
            event_manager.emit(mirror.event.TaskStopEvent(task.name, task.pid, task.code))
            task.set_stop_flag()
            return
        if task.stage == 1:
            log.info("Task: %s scheduled to second stage", task.name)
            self.run_task(TaskInfo(task.name, REGULAR_TASK, 0, task.priority), stage = 2)
        else:
            event_manager.emit(mirror.event.TaskStopEvent(task.name, task.pid, task.code))
            task.set_stop_flag()
            task.stage = 1

    def stop_all_tasks(self, signo = signal.SIGTERM):
        """
        This method can only be called when mirrord is shut down by SIGTERM or SIGINT.

        NOTE:
        Currently when mirrord is shut down, all running tasks will also be killed.

        """
        event_manager = component.get("EventManager")
        for taskname, task in self.tasks.iteritems():
            if task.isinternal:
                continue
            if not task.running:
                continue
            pid = task.pid
            task.stop(signo)
            # Not sure it is ok...
            pid, status  = os.waitpid(pid, 0)

            endstr, code = self.parse_return_status(status)
            task.code    = code
            event_manager.emit(mirror.event.TaskStopEvent(task.name, task.pid, task.code))
            log.info("Killed task: %s with pid %d", taskname, pid)

    @classmethod
    def get_runnable_priority(cls, current, limit):
        """
        If limit is zero, all priority tasks can be run.
        Else if current value is lower than limit, all priority tasks can be run.
        Else it is a function between target priority and (current / limit).

        """
        if limit <= 0:
            return PRIORITY_MAX
        if current < limit:
            return PRIORITY_MAX
        return (-4.55 * (current * 1.0 / limit)) + 14.55

    @classmethod
    def parse_return_status(cls, status):
        if (status & 0xff) != 0:
            endstr = "killed by signal"
            code   = (status & 0xff)
        else:
            endstr = "ended with return code"
            # See "EXIT VALUES" section in man rsync
            code   = (status >> 8)
        return (endstr, code)



#
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# mirror is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mirror. If not, write to:
# 	The Free Software Foundation, Inc.,
# 	51 Franklin Street, Fifth Floor
# 	Boston, MA  02110-1301, USA.
#
#

import logging
import shutil
import os

import sys
if sys.version_info.major >= 3:
    from configparser import ConfigParser
else:
    from ConfigParser import ConfigParser

import mirror.common

log = logging.getLogger(__name__)

def prop(func):
    """Function decorator for defining property attributes

    The decorated function is expected to return a dictionary
    containing one or more of the following pairs:
        fget - function for getting attribute value
        fset - function for setting attribute value
        fdel - function for deleting attribute
    This can be conveniently constructed by the locals() builtin
    function; see:
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/205183
    """
    return property(doc=func.__doc__, **func())

class Config(object):
    def __init__(self, filename, defaults=None, config_dir=None):
        self.__config           = {}
        self.__set_functions    = {}
        self.__change_callbacks = []

        if defaults:
            for key, value in defaults.iteritems():
                self.set_item(key, value)

        # Load the config from file in the config_dir
        if config_dir:
            self.__config_file = os.path.join(config_dir, filename)
        else:
            self.__config_file = mirror.common.get_default_config_dir(filename)

        self.load()

    def __iter__(self):
        for key in self.__config:
            yield key

    def __contains__(self, item):
        return item in self.__config

    def __setitem__(self, key, value):
        return self.set_item(key, value)

    def set_item(self, key, value):
        """
        **Usage**

        >>> config = Config("mirror.ini")
        >>> config["archlinux"] = {"upstream": "archlinux.org"}
        >>> config["archlinux"]
        {"upstream": "archlinux.org"}

        """
        if isinstance(value, basestring):
            value = mirror.common.utf8_encoded(value)

        if not self.__config.has_key(key):
            self.__config[key] = value
            log.debug("Setting '%s' to %s with type %s", key, value, type(value))
            return

        if self.__config[key] == value:
            return

        # Do not allow the type to change unless it is None
        oldtype, newtype = type(self.__config[key]), type(value)

        if value is not None and oldtype != type(None) and oldtype != newtype:
            try:
                if oldtype == unicode:
                    value = oldtype(value, "utf8")
                else:
                    value = oldtype(value)
            except ValueError:
                log.warning("Type '%s' invalid for '%s'", newtype, key)
                raise

        log.debug("Setting '%s' to %s with type %s", key, value, type(value))

        self.__config[key] = value

        # call set functions
        try:
            for func in self.__set_functions[key]:
                func(key, value)
        except KeyError:
            pass
        # call change callbacks
        try:
            for func in self.__change_callbacks:
                func(key, value)
        except:
            pass

    def __getitem__(self, key):
        return self.get_item(key)

    def get_item(self, key):
        """
        **Usage**

        >>> config = Config("mirror.ini")
        >>> config["archlinux"]
        {"upstream": "archlinux.org"}

        """
        if isinstance(self.__config[key], str):
            try:
                return self.__config[key].decode("utf8")
            except UnicodeDecodeError:
                return self.__config[key]
        else:
            return self.__config[key]

    def __delitem__(self, key):
        """
        See
        :meth:`del_item`
        """
        self.del_item(key)

    def del_item(self, key):
        """
        Deletes item with a specific key from the configuration.

        **Usage**
        >>> config = Config("mirror.ini")
        >>> del config["archlinux"]
        """
        del self.__config[key]

    def register_change_callback(self, callback):
        """
        **Usage**

        >>> config = Config("mirror.ini", defaults={"test": 5})
        >>> def callback(key, value):
        ...     print key, value
        ...
        >>> config.register_change_callback(callback)

        """
        self.__change_callbacks.append(callback)

    def register_set_function(self, key, function, apply_now=True):
        """
        Register a function to be called when a config value changes

        **Usage**

        >>> config = Config("mirror.ini", defaults={"test": 5})
        >>> def callback(key, value):
        ...     print key, value
        ...
        >>> config.register_set_function("test", callback, apply_now=True)
        test 5

        """
        log.debug("Registering function for %s key..", key)
        if key not in self.__set_functions:
            self.__set_functions[key] = []

        self.__set_functions[key].append(function)

        # Run the function now if apply_now is set
        if apply_now:
            function(key, self.__config[key])
        return

    def apply_all(self):
        """
        Calls all set functions

        **Usage**

        >>> config = Config("mirror.ini", defaults={"test": 5})
        >>> def callbck(key, value):
        ...     print key, value
        ...
        >>> config.register_set_function("test", callback, apply_now=False)
        >>> config.apply_all()
        test 5

        """
        log.debug("Calling all set functions..")
        for key, value in self.__set_functions.iteritems():
            for func in value:
                func(key, self.__config[key])

    def apply_set_functions(self, key):
        """
        Calls set functions for `:param:key`.

        """
        log.debug("Calling set functions for key %s..", key)
        if key in self.__set_functions:
            for func in self.__set_functions[key]:
                func(key, self.__config[key])

    def load(self, filename=None):
        """
        Load a config file

        """
        if not filename:
            filename = self.__config_file

        try:
            data = open(filename, "rb").read()
        except IOError as e:
            log.warning("Unable to open config file %s: %s", filename, e)
            return

        # load mirror ini config file
        config = ConfigParser()
        config.read(filename)
        for section in config.sections():
            value = {}
            for item in config.items(section):
                if item[0].endswith("[]") and item[0] not in value:
                    value[item[0]] = [item[1]]
                    continue
                if item[0] in value and type(value[item[0]]) == list:
                    value[item[0]].append(item[1])
                else:
                    value[item[0]] = item[1]
            self.set_item(section, value)

    def save(self, filename=None):
        """
        Save configuration to disk

        """
        if not filename:
            filename = self.__config_file

        # Save the new config and make sure it's written to disk
        try:
            log.debug("Saving new config file %s", filename + ".new")
            f = open(filename + ".new", "w")
            # TODO: save ini config
            for section, value in self.__config.items():
                f.write("[" + section + "]\n")
                for key, val in value.items():
                    if key.endswith("[]"):
                        for v in val:
                            f.write(key + "\t=\t" + v + "\n")
                    else:
                        f.write(key + "\t=\t" + val + "\n")
                f.write("\n")
            f.flush()
            os.fsync(f.fileno())
            f.close()
        except IOError as e:
            log.error("Error writing new config file: %s", e)
            return False

        # Make a backup of the old config
        try:
            log.debug("Backing up old config file to %s~", filename)
            shutil.move(filename, filename + "~")
        except Exception as e:
            log.warning("Unable to backup old config...")

        # The new config file has been written successfully, so let's move it over
        # the existing one.
        try:
            log.debug("Moving new config file %s to %s..", filename + ".new", filename)
            shutil.move(filename + ".new", filename)
        except Exception as e:
            log.error("Error moving new config file: %s", e)
            return False
        else:
            return True

    @property
    def config_file(self):
        return self.__config_file

    @prop
    def config():
        """The config dictionary"""
        def fget(self):
            return self.__config
        def fdel(self):
            return self.save()
        return locals()



#
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# mirror is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mirror. If not, write to:
#   The Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor
#   Boston, MA  02110-1301, USA.
#
#

events = {}

class MirrorEventMetaClass(type):
    """
    The metaclass keeps a list of all known event classes.

    """
    def __init__(cls, name, bases, dct):
        super(MirrorEventMetaClass, cls).__init__(name, bases, dct)
        if name != "MirrorEvent":
            events[name] = cls

class MirrorEvent(object):
    """
    The base class for all events.

    """
    # MirrorEventMetaClass is used to generate this class
    __metaclass__ = MirrorEventMetaClass

    def _get_name(self):
        return self.__class__.__name__

    def _get_args(self):
        if hasattr(self, "_args"):
            return self._args
        return []

    name = property(fget=_get_name)
    args = property(fget=_get_args)

class MirrorStartEvent(MirrorEvent):
    """
    The event occurs when mirror daemon is to be running.

    """
    pass

class TaskEnqueueEvent(MirrorEvent):
    """
    The event occurs when a task is appended into the queue.

    :param taskname: task's name

    """
    def __init__(self, taskname):
        self._args = [ taskname ]

class PreTaskStartEvent(MirrorEvent):
    """
    The event occurs when a new task is going to be running.

    :param taskname: task's name

    """
    def __init__(self, taskname):
        self._args = [ taskname ]

class TaskStartEvent(MirrorEvent):
    """
    The event occurs when a new task is running.

    :param taskname: task's name
    :param taskpid: task's process id

    """
    def __init__(self, taskname, taskpid):
        self._args = [ taskname, taskpid ]

class TaskStopEvent(MirrorEvent):
    """
    The event occurs when a task is ended.

    :param taskname: task's name
    :param taskpid: task's process id
    :param exitcode: task's exit code

    """
    def __init__(self, taskname, taskpid, exitcode):
        self._args = [ taskname, taskpid, exitcode ]

class RunSystemTaskEvent(MirrorEvent):
    """
    The event occurs when an internal task needs to run.

    :param taskinfo: TaskInfo instance

    """
    def __init__(self, taskinfo):
        self._args = [ taskinfo ]




#
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# mirror is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mirror. If not, write to:
#   The Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor
#   Boston, MA  02110-1301, USA.
#
#

import logging

log = logging.getLogger(__name__)

class PluginInitBase(object):

    _plugin_class = None

    def __init__(self, plugin_name):
        self.plugin = self._plugin_class(plugin_name)

    def enable(self):
        try:
            self.plugin.enable()
        except Exception as e:
            log.error("Unable to enable plugin: %s", self.plugin.name)
            log.exception(e)

    def disable(self):
        try:
            self.plugin.disable()
        except Exception as e:
            log.error("Unable to disable plugin: %s", self.plugin.name)
            log.exception(e)



#
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# mirror is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mirror. If not, write to:
#   The Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor
#   Boston, MA  02110-1301, USA.
#
#

MIRROR_OK     = 0
MIRROR_ERROR  = 1
MIRROR_ERRARG = 2

class MirrorError(Exception):
    def __new__(cls, *args, **kwargs):
        inst = super(MirrorError, cls).__new__(cls, *args, **kwargs)
        inst._args = args
        inst._kwargs = kwargs
        return inst

    def __init__(self, *args, **kwargs):
        if len(args) == 0:
            self._message = ""
        elif len(args) == 1:
            self._message = args[0]
        else:
            self._message = args[0] % args[1:]

    def _set_message(self, message):
        self._message = message

    def _get_message(self):
        return self._message

    message = property(_get_message, _set_message)

    del _get_message, _set_message

    def __str__(self):
        return self.message

class MirrordRunningError(MirrorError):
    pass




#!/usr/bin/env python
#
# setup.py
#
# Copyright (C) 2013 Shang Yuanchun <idealities@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, write to:
#   The Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor
#   Boston, MA    02110-1301, USA.
#

"""
Mirror is an open source python application for mirror site (e.g. mirror.bjtu.edu.cn)
to sync files from upstreams (it uses rsync internally), it actually works like a 
cron, but still has some differences. It has been served for mirror.bjtu.edu.cn with
more than 40 rsync tasks.
"""

try:
    from setuptools import setup, find_packages, Extension
except ImportError:
    import ez_setup
    ez_setup.use_setuptools()
    from setuptools import setup, find_packages, Extension

import os, sys
import glob
import shlex
import msgfmt
import platform
import unittest

from distutils import cmd
from distutils.command.build import build as _build
from distutils.command.clean import clean as _clean

python_version = platform.python_version()[0:3]

class build_trans(cmd.Command):
    description = 'Compile .po files into .mo file'

    user_options = [
            ('build-lib=', None, "lib build folder"),
            ('develop-mode', 'D', 'Compile translations in develop mode(into mirror/i18n)')
    ]
    boolean_options = ['develop_mode']

    def initialize_options(self):
        self.build_lib = None
        self.develop_mode = False

    def finalize_options(self):
        self.set_undefined_options('build', ('build_lib', 'build_lib'))

    def run(self):
        po_dir = os.path.join(os.path.dirname(__file__), 'mirror/i18n/')

        if self.develop_mode:
            basedir = po_dir
        else:
            basedir = os.path.join(self.build_lib, 'mirror', 'i18n')

        print('Compiling po files from %s...' % po_dir),
        for path, names, filenames in os.walk(po_dir):
            for f in filenames:
                uptodate = False
                if f.endswith('.po'):
                    lang = f[:len(f) - 3]
                    src = os.path.join(path, f)
                    dest_path = os.path.join(basedir, lang, 'LC_MESSAGES')
                    dest = os.path.join(dest_path, 'mirror.mo')
                    if not os.path.exists(dest_path):
                        os.makedirs(dest_path)
                    if not os.path.exists(dest):
                        sys.stdout.write('%s, ' % lang)
                        sys.stdout.flush()
                        msgfmt.make(src, dest)
                    else:
                        src_mtime = os.stat(src)[8]
                        dest_mtime = os.stat(dest)[8]
                        if src_mtime > dest_mtime:
                            sys.stdout.write('%s, ' % lang)
                            sys.stdout.flush()
                            msgfmt.make(src, dest)
                        else:
                            uptodate = True
                if uptodate:
                    sys.stdout.write(' po files already up to date.  ')
        sys.stdout.write('\b\b \nFinished compiling translation files. \n')

class build_plugins(cmd.Command):
    description = "Build plugins into .eggs"

    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        # Build the plugin eggs
        PLUGIN_PATH = "mirror/plugins/*"

        for path in glob.glob(PLUGIN_PATH):
            if os.path.exists(os.path.join(path, "setup.py")):
                os.system("cd " + path + " && " + sys.executable + " setup.py bdist_egg -d ..")


class build(_build):
    sub_commands = [ ('build_trans', None), ('build_plugins', None) ] + _build.sub_commands
    def run(self):
        # Run all sub-commands (at least those that need to be run)
        _build.run(self)

class clean(_clean):
    sub_commands = _clean.sub_commands

    def run(self):
        # Run all sub-commands (at least those that need to be run)
        for cmd_name in self.get_sub_commands():
            self.run_command(cmd_name)
        _clean.run(self)

class test(cmd.Command):
    """Command for running unittests without install."""

    user_options = [("args=", None, '''The command args string passed to
                                       unittest framework, such as
                                       --args="-v -f"''')]

    def initialize_options(self):
        self.args = ''

    def finalize_options(self):
        pass

    def run(self):
        self.run_command('build')
        bld = self.distribution.get_command_obj('build')
        # Add build_lib in to sys.path so that unittest can found DLLs and libs
        sys.path.insert(0, os.path.abspath(bld.build_lib))

        test_argv0 = [sys.argv[0] + ' test --args=']
        # For transfering args to unittest, we have to split args by ourself,
        # so that command like:
        #
        #   python setup.py test --args="-v -f"
        #
        # can be executed, and the parameter '-v -f' can be transfering to
        # unittest properly.
        test_argv = test_argv0 + shlex.split(self.args)
        unittest.main(None, defaultTest='test.test_suite', argv=test_argv)

cmdclass = {
    'build': build,
    'build_trans': build_trans,
    'build_plugins': build_plugins,
    'test': test,
    'clean': clean,
}

# Data files to be installed to the system
_data_files = [
    ('share/man/man1',   ['docs/man/mirrord.1']),
    ('share/mirror/etc', ['config/mirror.ini'])
]

entry_points = {
    "console_scripts": [
        "mirrord = mirror.main:start_daemon"
    ],
}

# Main setup
setup(
    name = "mirror",
    version = "0.6.5",
    fullname = "Mirror for open source mirror site",
    description = "Mirror for open source mirror site to sync files",
    author = "Shang Yuanchun",
    author_email = "idealities@gmail.com",
    keywords = "open source mirror",
    long_description = __doc__,
    url = "http://mirror.bjtu.edu.cn",
    license = "GPLv3",
    cmdclass = cmdclass,
    data_files = _data_files,
    package_data = {"mirror": [
                               "plugins/*.egg",
                               "i18n/*/LC_MESSAGES/*.mo",
                               "README.md",
                               "msgfmt.py",
                               "AUTHORS",
                               "LICENSE",
                               ]},
    packages = find_packages(exclude=["docs","test"]),
    entry_points = entry_points
)



#
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# mirror is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mirror. If not, write to:
#   The Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor
#   Boston, MA  02110-1301, USA.
#
#


"""Common functions for Mirror :("""

import os, sys
import re
import time
import logging
import pkg_resources
import gettext
import locale
try:
    import chardet
except:
    pass

from mirror.error import *

log = logging.getLogger(__name__)

DEFAULT_MIRRORD_LOG_DIR = '/var/log/mirrord'
DEFAULT_TASK_LOG_DIR    = '/var/log/rsync'

def get_version():
    """
    Returns the version of mirror from the python egg metadata

    :returns: the version of mirror

    """
    return pkg_resources.require("mirror")[0].version

def get_default_config_dir(filename=None):
    """
    :param filename: if None, only the config directory path is returned,
                     if provided, a path including the filename will be returned
    :type  filename: string
    :returns: a file path to the config directory and optional filename
    :rtype: string

    """

    try:
        from xdg.BaseDirectory import save_config_path
        config_path = save_config_path("mirror")
    except:
        config_path = os.path.join(os.path.expanduser("~/.config"),
                                   "mirror")
    if not filename:
        filename = ''
    return os.path.join(config_path, filename)

def setup_translations():
    translations_path = resource_filename("mirror", "i18n")
    log.info("Setting up translations from %s", translations_path)

    try:
        if hasattr(locale, "bindtextdomain"):
            locale.bindtextdomain("mirror", translations_path)
        if hasattr(locale, "textdomain"):
            locale.textdomain("mirror")
        gettext.install("mirror", translations_path, unicode=True)
    except Exception as e:
        log.error("Unable to initialize gettext/locale")
        log.exception(e)
        import __builtin__
        __builtin__.__dict__["_"] = lambda x: x

def resource_filename(module, path):
    return pkg_resources.require(
               "mirror>=%s" % get_version())[0].get_resource_filename(
                pkg_resources._manager, os.path.join(*(module.split('.')+[path]))
            )

def read_mirrord_pid(pidfile):
    pid = None
    if os.path.isfile(pidfile):
        try:
            pid = int(open(pidfile).read().strip())
        except:
            pass
    return pid

def check_mirrord_running(pidfile):
    pid = read_mirrord_pid(pidfile)

    def is_process_running(pid):
        try:
            os.kill(pid, 0)
        except OSError:
            return False
        else:
            return True

    if pid and is_process_running(pid):
        raise MirrordRunningError("Another mirrord is running with pid: %d", pid)

def lock_file(pidfile):
    """
    Actually the code below is needless...

    """

    import fcntl
    try:
        fp = open(pidfile, "r+" if os.path.isfile(pidfile) else "w+")
    except IOError:
        raise MirrorError("Can't open or create %s", pidfile)

    try:
        fcntl.flock(fp, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except IOError:
        try:
            pid = int(fp.read().strip())
        except:
            raise MirrorError("Can't lock %s", pidfile)
        raise MirrorError("Can't lock %s, maybe another mirrord with pid %d is running",
                          pidfile, pid)

    # See document at http://man7.org/linux/man-pages/man2/fcntl.2.html
    flag = fcntl.fcntl(fp, fcntl.F_GETFD)
    fcntl.fcntl(fp, fcntl.F_SETFD, flag | fcntl.FD_CLOEXEC)

    fp.seek(0)
    fp.write("%d\n" % os.getpid())
    fp.truncate()
    fp.flush()

    # We need to return fp to keep a reference on it
    return fp

def find_command(command):
    """
    Find the path of `command`.

    :returns: the path of command or None if not found

    """
    paths = os.getenv("PATH").split(":")
    for path in paths:
        filepath = (path if path.endswith('/') else path + '/') + command
        if os.path.isfile(filepath):
            return filepath
    return None

def parse_timeout(timeout):
    """
    Parse timeout expression, e.g. 12h17m, 12h, 17m

    :returns: the seconds represented by timeout, or 0 if timeout is not valid

    """
    try:
        return int(timeout)
    except:
        pass
    h = timeout.find('h')
    m = timeout.find('m')
    if h > 0 or m > 0:
        try:
            return ((int(timeout[:h]) * 3600 if h > 0 else 0)
                   + (int(timeout[h+1:m]) * 60 if m > 0 else 0))
        except:
            return 0
    else:
        return 0

CRON_TIME = re.compile(r'^\s*([^@#\s]+)\s+([^@#\s]+)\s+([^@#\s]+)' +
                       r'\s+([^@#\s]+)\s+([^@#\s]+)\s*(#\s*([^\n]*)|$)')
CRON_ITEM = re.compile(r'^(\d+)-(\d+)/(\d+)$')

def parse_cron_time(time):
    """
    Parse the cron time format, e.g. */20 * * * *

    :returns: a tuple with 7 elements,
              (minute, hour, day of month, month, day of week, comment with #,
               comment).
              Or None if `time` is not valid
    """
    extent      = (60, 24, 32, 13, 8)
    result_text = CRON_TIME.findall(time)
    if result_text:
        result_text = result_text[0]
        result  = []
        for i in xrange(5):
            value = result_text[i]
            items = CRON_ITEM.findall(value)
            if value == '*':
                result.append([d for d in xrange(i >= 2, extent[i])])
            elif value.startswith('*/'):
                every = int(value.split('/')[1])
                result.append([d for d in xrange(i >= 2, extent[i], every)])
            elif items:
                item  = items[0]
                start = int(item[0])
                end   = int(item[1])
                every = int(item[2])
                if start > extent[i]:
                    start = extent[i]
                if start < (i >= 2):
                    start = (i >= 2)
                if end   > extent[i]:
                    end   = extent[i]
                result.append([d for d in xrange(start, end, every)])
            elif value.find(',') != -1:
                result.append(
                        sorted([int(d) for d in value.split(',')
                               if (int(d) < extent[i] and int(d) >= (i >= 2))])
                       )
            else:
                try:
                    d = int(value)
                    if (int(value) < extent[i] and int(value) >= (i >= 2)):
                        result.append([d])
                    else:
                        result.append([])
                except:
                    result.append([])
        result += result_text[-2:]
        return result
    else:
        return None

def write_stderr(message, *args):
    """
    """
    sys.stderr.write(message % args)
    sys.stderr.write("\n")

def decode_string(s, encoding="utf8"):
    """
    Decodes a string and return unicode. If it cannot decode using
    `:param:encoding` then it will try latin1, and if that fails,
    try to detect the string encoding. If that fails, decode with
    ignore.

    """
    if not s:
        return u''
    elif isinstance(s, unicode):
        return s

    encodings = [lambda: ("utf8", 'strict'),
                 lambda: ("iso-8859-1", 'strict'),
                 lambda: (chardet.detect(s)["encoding"], 'strict') if
                         'chardet' in sys.modules else (encoding, 'ignore'),
                 lambda: (encoding, 'ignore')]

    if encoding != "utf8":
        encodings.insert(0, lambda: (encoding, 'strict'))

    for e in encodings:
        try:
            return s.decode(*e())
        except UnicodeDecodeError:
            pass
    return u''

def utf8_encoded(s, encoding="utf8"):
    """
    Return a utf8 encoded string of s

    """
    if isinstance(s, str):
        s = decode_string(s, encoding).encode("utf8")
    elif isinstance(s, unicode):
        return s.encode('utf8')

    return s

def is_python3():
    """
    """
    return sys.version_info.major >= 3



#
# Copyright (C) 2013 Shang Yuanchun <idealities@gmail.com>
#
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# mirror is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mirror. If not, write to:
#   The Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor
#   Boston, MA  02110-1301, USA.
#
#

import os, sys
import time
import mmap
import struct
import signal
import cPickle as pickle
import mirror.color
import mirror.common
import mirror.task  as task
import mirror.error as error
from mirror.common import write_stderr

TASK_DESC = {
            task.REGULAR_TASK: "Normal task",
            task.TIMEOUT_TASK: "Timeout check",
            task.SYSTEM_TASK : "System task",
            }

def list_task_queue():
    try:
        bufferfd = os.open("/tmp/mirrord", os.O_RDONLY)
    except:
        write_stderr(_("Open /tmp/mirrord failed, "
                       "can't read task infomation"))
        return error.MIRROR_ERROR

    buffer = mmap.mmap(bufferfd, os.fstat(bufferfd).st_size,
                       mmap.MAP_SHARED, mmap.PROT_READ)
    os.close(bufferfd)
    if buffer[:2] != '\x79\x71':
        write_stderr(_("Wrong file /tmp/mirrord, "
                       "any other writed it?"))
        return error.MIRROR_ERROR

    buffer.seek(2)
    size = struct.unpack("I", buffer.read(4))[0]

    taskqueue = pickle.loads(buffer.read(size))
    formatstr = ("Task:"+mirror.color.FOREGROUND_COLORS.GREEN
                +"%-18s"+mirror.color.COLOR_END+"\ttype:%14s\ttime: %s")
    for taskinfo in taskqueue:
        print formatstr % (
              taskinfo.name, TASK_DESC[taskinfo.tasktype],
              time.asctime(time.localtime(taskinfo.time)))

    buffer.close()
    return error.MIRROR_OK

signals = {
          "stop"  : signal.SIGQUIT,
          "reload": signal.SIGHUP,
          }

def signal_process(signame):
    signames = signals.keys()
    if signame not in signames:
        write_stderr(_("Invalid value for -s, "
                       "available: ") + ", ".join(signames))
        return error.MIRROR_ERRARG
    import mirror.configmanager
    pidfile = mirror.configmanager.get_config_dir("mirrord.pid")
    pid = mirror.common.read_mirrord_pid(pidfile)
    if not pid:
        write_stderr(_("Can not read mirrord's pid, "
                       "please make sure that mirrord is running"))
        return error.MIRROR_ERROR
    try:
        os.kill(pid, signals.get(signame))
    except Exception as e:
        write_stderr(_("Kill mirrord (%d) failed: %s"),
                     pid, e)
        return error.MIRROR_ERROR

    return error.MIRROR_OK




#
#
# You may redistribute it and/or modify it under the terms of the
# GNU General Public License, as published by the Free Software
# Foundation; either version 3 of the License, or (at your option)
# any later version.
#
# mirror is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with mirror. If not, write to:
#   The Free Software Foundation, Inc.,
#   51 Franklin Street, Fifth Floor
#   Boston, MA  02110-1301, USA.
#
#

import logging
from collections import defaultdict

log = logging.getLogger(__name__)

class Component(object):

    def __init__(self, name, depend = None):
        self._name   = name
        self._depend = depend
        _component_registry.register(self)

    def __del__(self):
        if _component_registry:
            _component_registry.deregister(self)

    def start(self):
        pass

    def stop(self):
        pass

class ComponentRegistry(object):

    def __init__(self):
        self.components = {}
        # Stores all of the components that are dependent on a
        # particular component
        self.dependents = defaultdict(list)

    def register(self, obj):
        """
        Register a component object with the registry.  This is done
        automatically when a Component object is instantiated.

        :param obj: the Component object

        :raises ComponentAlreadyRegistered: if a component with the
        same name is already registered.

        """
        name = obj._name

        if not isinstance(obj, Component):
            log.warn("Trying to register a object with type: %s",
                         obj.__class__.__name__)
            return

        if name in self.components:
            log.warn(("Trying to register a object which is already"
                         " registered, name: %s"), 
                         name)

        self.components[obj._name] = obj

        if obj._depend:
            for depend in obj._depend:
                self.dependents[depend].append(name)

    def deregister(self, obj):
        """
        Deregisters a component from the registry.

        :param obj: the Component object

        """

        if obj in self.components.values():
            log.debug("Deregistering Component: %s", obj._name)
            self.stop([obj._name])
            self.components.pop(obj._name)

    def start(self, names = []):
        """
        Starts Components that are currently in a Stopped state and their
        dependencies.  If *names* is specified, will only start those
        Components and their dependencies and if not it will start all
        registered components.

        :param names: a list of Components to start

        """
        # Start all the components if names is empty
        if not names:
            names = self.components.keys()
        elif isinstance(names, str):
            names = [ names ]

        for name in names:
            if self.components[name]._depend:
                # This component has depends, so we need to start them first.
                self.start(self.components[name]._depend)
            log.debug("Starting component: %s", name)
            self.components[name].start()

    def stop(self, names = []):
        """
        Stops Components that are currently not in a Stopped state.  If
        *names* is specified, then it will only stop those Components,
        and if not it will stop all the registered Components.

        :param names: a list of Components to stop

        """
        if not names:
            names = self.components.keys()
        elif isinstance(names, str):
            names = [ names ]

        stopped = set()

        for name in names:
            if name in stopped:
                continue
            if name not in self.components:
                continue
            if name in self.dependents:
                # If other components depend on this component, stop them first
                self.stop(self.dependents[name])
                stopped.update(self.dependents[name])
            self.components[name].stop()

_component_registry = ComponentRegistry()

# methods provided by mirror.component

def get(name):
    """
    Return a reference to a component.

    :param name: the Component name to get

    :returns: the Component object, or None
    if the Component does not exist

    """
    return _component_registry.components.get(name, None)

deregister = _component_registry.deregister
start      = _component_registry.start
stop       = _component_registry.stop

